'''
Created on 2022-04-21

@author: wf
'''

class Version(object):
    '''
    Version handling for pyGenericSpreadSheet
    '''
    version='0.0.23'
    date = '2022-04-21'
    updated = '2022-07-27'
    name='pyGenericSpreadSheet'
    description='python API providing generic Access to Spreadsheet backends'

        
